@@ -11,5 +11,6 @@ from .functional import (
     config_to_yaml,
     config_to_program
 )
-from . import defaults
-from .instantiate import instantitate
\ No newline at end of file
+from . import defaults as _defaults
+from .instantiate import instantitate
+from .logging import configure_logging
